package com.genericparallel;

import java.util.Map;

import javax.script.Bindings;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.w3c.dom.Node;

import com.bea.wli.config.Ref;
import com.bea.wli.config.component.NotFoundException;
import com.bea.wli.sb.resources.config.XmlEntryDocument;
import com.bea.wli.sb.resources.xml.XmlRepository;
import com.bea.wli.sb.resources.xquery.XqueryRepository;

public class Javascript {

	/**
	 * Execute a script with the body of JSON and return the result.
	 * @param json JSON data as string
	 * @param scriptPath Path to XML containing Javascript under OSB Projects (e.g. 'Foo/Bar/InsertCredentials.xml') 
	 * @param xml Other parameters for Javascript
	 * @return New JSON value.
	 */
	public static String execute(String json, String scriptPath, XmlObject xml) throws Exception {
	    // load JS from path
	    XmlObject scriptXml = readXml(scriptPath);
	    
	    XmlCursor cursor = scriptXml.newCursor();
	    cursor.toFirstChild();
	    String script = cursor.getTextValue();
	    if( script != null ) script = script.trim();

	    return executeInner(json,script,xml);
	}	
	    		
	private static String executeInner(String json, String script, XmlObject xml) throws Exception {
	    ScriptEngineManager manager = new ScriptEngineManager();
	    ScriptEngine engine = manager.getEngineByName("javascript");

	    // collect parameters
	    Bindings bindings = engine.createBindings();
	    if( xml != null ) {
	    	XmlObject[] params = xml.execQuery("//var");
	    	for (XmlObject p : params) {
	    		String name = null;
	    		String value = null;

	    	    XmlCursor cursor = p.newCursor();
	    	    cursor.toFirstChild();
	    	    cursor.toFirstChild();

	    	    String varName = cursor.getName().getLocalPart();
				if( "name".equals(varName) ) {
					name = cursor.getTextValue();
				}
				if( "value".equals(varName) ) {
					value = cursor.getTextValue();
				}
				
				cursor.toNextSibling();

	    	    varName = cursor.getName().getLocalPart();
				if( "name".equals(varName) ) {
					name = cursor.getTextValue();
				}
				if( "value".equals(varName) ) {
					value = cursor.getTextValue();
				}
				
				if( name != null ) {
					bindings.put(name, value);
				}
			}
	    }

	    addJSON(engine,bindings);
	    
	    // inject and parse the main value
	    engine.eval("var json = "+json+";",bindings);
	    
	    // JSONPath scripts just have $..ban format
	    if( script.startsWith("$") ) {
		    addJSONPath(engine,bindings);
		    
	    	// need to escape quotes in script, but I'm too lazy to do that for a demo
	    	script = "var output = jsonPath(json,\""+script+"\"); if( output === false ) output = null; else output = output[0];";
		    engine.eval(script,bindings);
		    return (String)engine.eval("typeof output=='string'||output==null?output:JSON.stringify(output)",bindings);
	    } else {
	    	// execute the script
	    	engine.eval(script,bindings);
	    	// return the value of output variable as String
	    	return (String)engine.eval("typeof output=='string'?output:JSON.stringify(output)",bindings);
	    }
	}
	
    private static void addJSONPath(ScriptEngine engine, Bindings bindings) throws ScriptException {
    	engine.eval(
    			"/* JSONPath 0.8.0 - XPath for JSON\n"+
				" *\n"+
				" * Copyright (c) 2007 Stefan Goessner (goessner.net)\n"+
				" * Licensed under the MIT (MIT-LICENSE.txt) licence.\n"+
				" */\n"+
				"function jsonPath(obj, expr, arg) {\n"+
				"   var P = {\n"+
				"      resultType: arg && arg.resultType || \"VALUE\",\n"+
				"      result: [],\n"+
				"      normalize: function(expr) {\n"+
				"         var subx = [];\n"+
				"         return expr.replace(/[\\['](\\??\\(.*?\\))[\\]']/g, function($0,$1){return \"[#\"+(subx.push($1)-1)+\"]\";})\n"+
				"                    .replace(/'?\\.'?|\\['?/g, \";\")\n"+
				"                    .replace(/;;;|;;/g, \";..;\")\n"+
				"                    .replace(/;$|'?\\]|'$/g, \"\")\n"+
				"                    .replace(/#([0-9]+)/g, function($0,$1){return subx[$1];});\n"+
				"      },\n"+
				"      asPath: function(path) {\n"+
				"         var x = path.split(\";\"), p = \"$\";\n"+
				"         for (var i=1,n=x.length; i<n; i++)\n"+
				"            p += /^[0-9*]+$/.test(x[i]) ? (\"[\"+x[i]+\"]\") : (\"['\"+x[i]+\"']\");\n"+
				"         return p;\n"+
				"      },\n"+
				"      store: function(p, v) {\n"+
				"         if (p) P.result[P.result.length] = P.resultType == \"PATH\" ? P.asPath(p) : v;\n"+
				"         return !!p;\n"+
				"      },\n"+
				"      trace: function(expr, val, path) {\n"+
				"         if (expr) {\n"+
				"            var x = expr.split(\";\"), loc = x.shift();\n"+
				"            x = x.join(\";\");\n"+
				"            if (val && val.hasOwnProperty(loc))\n"+
				"               P.trace(x, val[loc], path + \";\" + loc);\n"+
				"            else if (loc === \"*\")\n"+
				"               P.walk(loc, x, val, path, function(m,l,x,v,p) { P.trace(m+\";\"+x,v,p); });\n"+
				"            else if (loc === \"..\") {\n"+
				"               P.trace(x, val, path);\n"+
				"               P.walk(loc, x, val, path, function(m,l,x,v,p) { typeof v[m] === \"object\" && P.trace(\"..;\"+x,v[m],p+\";\"+m); });\n"+
				"            }\n"+
				"            else if (/,/.test(loc)) { // [name1,name2,...]\n"+
				"               for (var s=loc.split(/'?,'?/),i=0,n=s.length; i<n; i++)\n"+
				"                  P.trace(s[i]+\";\"+x, val, path);\n"+
				"            }\n"+
				"            else if (/^\\(.*?\\)$/.test(loc)) // [(expr)]\n"+
				"               P.trace(P.eval(loc, val, path.substr(path.lastIndexOf(\";\")+1))+\";\"+x, val, path);\n"+
				"            else if (/^\\?\\(.*?\\)$/.test(loc)) // [?(expr)]\n"+
				"               P.walk(loc, x, val, path, function(m,l,x,v,p) { if (P.eval(l.replace(/^\\?\\((.*?)\\)$/,\"$1\"),v[m],m)) P.trace(m+\";\"+x,v,p); });\n"+
				"            else if (/^(-?[0-9]*):(-?[0-9]*):?([0-9]*)$/.test(loc)) // [start:end:step]  phyton slice syntax\n"+
				"               P.slice(loc, x, val, path);\n"+
				"         }\n"+
				"         else\n"+
				"            P.store(path, val);\n"+
				"      },\n"+
				"      walk: function(loc, expr, val, path, f) {\n"+
				"         if (val instanceof Array) {\n"+
				"            for (var i=0,n=val.length; i<n; i++)\n"+
				"               if (i in val)\n"+
				"                  f(i,loc,expr,val,path);\n"+
				"         }\n"+
				"         else if (typeof val === \"object\") {\n"+
				"            for (var m in val)\n"+
				"               if (val.hasOwnProperty(m))\n"+
				"                  f(m,loc,expr,val,path);\n"+
				"         }\n"+
				"      },\n"+
				"      slice: function(loc, expr, val, path) {\n"+
				"         if (val instanceof Array) {\n"+
				"            var len=val.length, start=0, end=len, step=1;\n"+
				"            loc.replace(/^(-?[0-9]*):(-?[0-9]*):?(-?[0-9]*)$/g, function($0,$1,$2,$3){start=parseInt($1||start);end=parseInt($2||end);step=parseInt($3||step);});\n"+
				"            start = (start < 0) ? Math.max(0,start+len) : Math.min(len,start);\n"+
				"            end   = (end < 0)   ? Math.max(0,end+len)   : Math.min(len,end);\n"+
				"            for (var i=start; i<end; i+=step)\n"+
				"               P.trace(i+\";\"+expr, val, path);\n"+
				"         }\n"+
				"      },\n"+
				"      eval: function(x, _v, _vname) {\n"+
				"         try { return $ && _v && eval(x.replace(/@/g, \"_v\")); }\n"+
				"         catch(e) { throw new SyntaxError(\"jsonPath: \" + e.message + \": \" + x.replace(/@/g, \"_v\").replace(/\\^/g, \"_a\")); }\n"+
				"      }\n"+
				"   };\n"+
				"\n"+
				"   var $ = obj;\n"+
				"   if (expr && obj && (P.resultType == \"VALUE\" || P.resultType == \"PATH\")) {\n"+
				"      P.trace(P.normalize(expr).replace(/^\\$;/,\"\"), obj, \"$\");\n"+
				"      return P.result.length ? P.result : false;\n"+
				"   }\n"+
				"}\n"    			
    			,bindings);
    }
	
    private static void addJSON(ScriptEngine engine, Bindings bindings) throws ScriptException {
    	engine.eval(
    			"JSON = {\n"+
    			"    parse: function(sJSON) { return eval('(' + sJSON + ')'); },\n"+
    			"    stringify: (function () {\n"+
    			"      var toString = Object.prototype.toString;\n"+
    			"      var isArray = Array.isArray || function (a) { return toString.call(a) === '[object Array]'; };\n"+
    			"      var escMap = {'\\\"': '\\\\\"', '\\\\': '\\\\\\\\', '\\b': '\\\\b', '\\f': '\\\\f', '\\n': '\\\\n', '\\r': '\\\\r', '\\t': '\\\\t'};\n"+
    			"      var escMap = {};\n"+
    			"      var escFunc = function (m) { return escMap[m] || '\\u' + (m.charCodeAt(0) + 0x10000).toString(16).substr(1); };\n"+
    			"      var escRE = /[\\\"\\u0000-\\u001F\\u2028\\u2029]/g;\n"+
    			"      return function stringify(value) {\n"+
    			"        if (value == null) {\n"+
    			"          return 'null';\n"+
    			"        } else if (typeof value === 'number') {\n"+
    			"          return isFinite(value) ? value.toString() : 'null';\n"+
    			"        } else if (typeof value === 'boolean') {\n"+
    			"          return value.toString();\n"+
    			"        } else if (typeof value === 'object') {\n"+
    			"          if (typeof value.toJSON === 'function') {\n"+
    			"            return stringify(value.toJSON());\n"+
    			"          } else if (isArray(value)) {\n"+
    			"            var res = '[';\n"+
    			"            for (var i = 0; i < value.length; i++)\n"+
    			"              res += (i ? ', ' : '') + stringify(value[i]);\n"+
    			"            return res + ']';\n"+
    			"          } else if (toString.call(value) === '[object Object]') {\n"+
    			"            var tmp = [];\n"+
    			"            for (var k in value) {\n"+
    			"              if (value.hasOwnProperty(k))\n"+
    			"                tmp.push(stringify(k) + ': ' + stringify(value[k]));\n"+
    			"            }\n"+
    			"            return '{' + tmp.join(', ') + '}';\n"+
    			"          }\n"+
    			"        }\n"+
    			"        return '\"' + value.toString().replace(escRE, escFunc) + '\"';\n"+
    			"      };\n"+
    			"    })()\n"+
    			"  };",bindings);
	}

	private static XqueryRepository xqr = XqueryRepository.get();

    /**
	 * Reads XML resource as XmlObject
	 *  
	 * @param xmlRefPath Project path e.g. "Foo/XML/Data"
	 * @return XmlObject with parsed XML
	 */
    private static XmlObject readXml(String xmlRefPath) {
    	Ref ref = new Ref("XML", Ref.getNames(xmlRefPath));
        XmlObject xmlObject = null;
        try {
            XmlEntryDocument xmlEntryDocument = XmlRepository.get().getEntry(ref);
            String xmlContent = xmlEntryDocument.getXmlEntry().getXmlContent();
            xmlObject = XmlObject.Factory.parse(xmlContent);
        } catch (NotFoundException e) {
            throw new RuntimeException("XML Resource not found: "+xmlRefPath);
        } catch (XmlException e) {
            throw new RuntimeException("Error parsing XML content: "+xmlRefPath, e);
        }
        return xmlObject;
    }
}